/*
 This file sets up the raspberry pi pins for NKK functionality
 */
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/types.h>
#include <bcm2835.h>

void SetupPins(void);
void SetupSpi(void);

// Provides all setup needed for raspberry pi inputs and outputs
void SetupRaspberryPiIo(void)
{
	bcm2835_init();
	
	SetupPins();
	SetupSpi();
}

// Set the needed pins as inputs or outputs
void SetupPins(void)
{
	// Inputs
	bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_03, BCM2835_GPIO_FSEL_INPT); // Switch 1 pressed (active low)
	
	// Outputs
    bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_11, BCM2835_GPIO_FSEL_OUTP); // Shutdown pin
	bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_13, BCM2835_GPIO_FSEL_OUTP); // Data/Command
	bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_15, BCM2835_GPIO_FSEL_OUTP); // Switch Select
	bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_16, BCM2835_GPIO_FSEL_OUTP); // Switch Reset
	//bcm2835_gpio_fsel(RPI_BPLUS_GPIO_J8_40, BCM2835_GPIO_FSEL_OUTP); // 
}

// Sets up the specific pins we use for SPI
// SPI is set up for specific pins that cannot be changed. They are:
// MOSI P1-19
// MISO P1-21
// SCLK P1-23
void SetupSpi(void)
{
	bcm2835_spi_begin();
	
	bcm2835_spi_setBitOrder(BCM2835_SPI_BIT_ORDER_MSBFIRST);      // The default
    bcm2835_spi_setDataMode(BCM2835_SPI_MODE0);                   // The default
    bcm2835_spi_setClockDivider(BCM2835_SPI_CLOCK_DIVIDER_8); // The default
    bcm2835_spi_chipSelect(BCM2835_SPI_CS_NONE);                      // The default
    //bcm2835_spi_setChipSelectPolarity(BCM2835_SPI_CS_NONE, LOW);      // the default
}

// Safely resets the pins
void SafelyCloseIoFunctions(void)
{
	bcm2835_close();
}


// Enables power to the switches
void EnablePower(void)
{
	bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_11, HIGH);
}

// Active low. Takes all switches out of reset mode
void DisableSwitchReset(void)
{
	bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_16, HIGH);
}

// Sets the Data/Command line low, which enables commands and disables image data
void EnableSwitchCommandSending(void)
{
	bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_13, LOW);
}

// Sets the Data/Command line high, which disables commands and enables image data
void DisableSwitchCommandSending(void)
{
	bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_13, HIGH);
}

// Active low. Selects switch 1 SPI chip select
void EnableSwitch1ChipSelect(void)
{
    bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_15, LOW);
}

// Active low. De-selects switch 1 SPI chip select
void DisableSwitch1ChipSelect(void)
{
    bcm2835_gpio_write(RPI_BPLUS_GPIO_J8_15, HIGH);
}

// Active low. True if switch is pressed, false otherwise
bool ReadSwitchPress(void)
{
	bool activeLow = bcm2835_gpio_lev(RPI_BPLUS_GPIO_J8_03);
	
	bool pressed = !activeLow;
	
	return pressed;
}