#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <sys/types.h>
#include <bcm2835.h>
#include "SwitchFunctions.h"
#include "IoFunctions.h"

#define STANDARD_SWITCH_IMAGE_SIZE 6144 // 64 * 48 * 2 bytes per pixel
#define FRAMELESS_SWITCH_IMAGE_SIZE 12288 // 96 * 64 * 2 bytes per pixel
#define WEBCAM_ROW_SIZE 1280 // 640 * 2 bytes per pixel (for color)
#define WEBCAM_TOTAL_SIZE 614400
#define FRAMELESS_IMAGE_SCALING_ROW 51200

void ConfigureFramelessOledSwitch(void);
void ConfigureStandardOledSwitch(void);
void DisableSwitchReset(void);
void EnablePower(void);
void LoadBlankImageOnSwitch(void);
void LoadImageOnSwitch(void);
bool ReadSwitchPress(void);
void LoadStandardImage(char * inputImage);
void LoadFramelessImage(char * inputImage);

unsigned char outputData[STANDARD_SWITCH_IMAGE_SIZE]; // Holds the 64.48 image to send to the switch

// CHANGE THIS WHEN CHANGING SWITCH TYPES
switchTypeEnum GetCurrentSwitchType(void)
{
	return SWITCH_TYPE_FRAMELESS;
}

// Does all setup needed for the switch to function properly
void SetupSwitch(void)
{
	EnablePower();
	DisableSwitchReset();
	
	switchTypeEnum type = GetCurrentSwitchType();		
	switch(type)
	{
		default:
		case SWITCH_TYPE_FRAMELESS:
			ConfigureFramelessOledSwitch();
			break;
		case SWITCH_TYPE_STANDARD:
			ConfigureStandardOledSwitch(); 
			break;
	}		
	
	LoadBlankImageOnSwitch();
}

// Sends a command and its parameters to both switches
void SendSwitchConfigurationCommand(unsigned char command, unsigned char* parameters, unsigned char numberOfParameters)
{
	unsigned char i;
	
	EnableSwitch1ChipSelect();
    EnableSwitchCommandSending();
    
    bcm2835_spi_transfer(command);
    
    for (i = 0; i < numberOfParameters; i++)
    {
        bcm2835_spi_transfer(parameters[i]);
    }
    
    DisableSwitchCommandSending();
	DisableSwitch1ChipSelect();
}

// Loads the image at the selected address (0001, 0002, etc.) on switch 1
void LoadBlankImageOnSwitch(void)
{
	unsigned int i;
	unsigned int size;
    EnableSwitch1ChipSelect();
	
	switchTypeEnum type = GetCurrentSwitchType();		
	switch(type)
	{
		default:
		case SWITCH_TYPE_FRAMELESS:
			size = FRAMELESS_SWITCH_IMAGE_SIZE;
			break;
		case SWITCH_TYPE_STANDARD:
			size = STANDARD_SWITCH_IMAGE_SIZE;
			break;
	}	
		
	for (i = 0; i < size; i++)
	{
		bcm2835_spi_transfer(0xFF);
	}
    
    DisableSwitch1ChipSelect();    
}

// The Logitech c200 webcam provideds video as YUYV color, which can also be interpreted
// as black and white. We convert it to RGB the switch needs, using the formulas found at:
// https://en.wikipedia.org/wiki/YUV
void LoadWebcamImageOnSwitch(char * inputImage)
{
	switchTypeEnum type = GetCurrentSwitchType();
	
	if (inputImage == 0) { return; } // Error catching
		
	switch(type)
	{
		default:
		case SWITCH_TYPE_FRAMELESS:
			LoadFramelessImage(inputImage);
			break;
		case SWITCH_TYPE_STANDARD:
			LoadStandardImage(inputImage);
			break;
	}		
}

// Gets the RGB from the specified index in the buffer. NO ERROR CHECKING
void GetSwitchRgbBytesFromYuyvAndSend(char * inputImage, unsigned int index)
{
		unsigned char redByte, blueByte, greenByte;
		int cr,cb,y0,y1;
		char spiBuffer[2];
	
		y0 = (char)inputImage[index];
		cb = (char)inputImage[index + 1] - 128;
		y1 = (char)inputImage[index + 2]; // y1 is for the 2nd pixel, which we don't use
		cr = (char)inputImage[index + 3] - 128; 
		
		spiBuffer[0] = 0;
		spiBuffer[1] = 0;
									
		blueByte = y0 + cb + (cb >> 1) + (cb >> 2) + (cb >> 6);
		greenByte = y0 - ((cb >> 2) + (cb >> 4) + (cb >> 5)) - ((cr >> 1) + (cr >> 3) + (cr >> 4) + (cr >> 5));			
		redByte = y0 + cr + (cr >> 2) + (cr >> 3) + (cr >> 5);
			   
		spiBuffer[0] |= (blueByte & 0xF8);
		spiBuffer[0] |= ((greenByte & 0xE0) >> 5); // The high byte
		
		spiBuffer[1] |= ((greenByte & 0x1C) << 3);
		spiBuffer[1] |= ((redByte & 0xF8) >> 3); // The low byte	
		
		bcm2835_spi_writenb(spiBuffer, 2);
}

// Gets bytes from a 640x480 image and converts to switch size
void LoadStandardImage(char * inputImage)
{
	unsigned int i;
	
	DisableSwitchCommandSending(); // When commands are disabled data is able to be sent    
	EnableSwitch1ChipSelect();		
	
	for (i = 0; i < WEBCAM_TOTAL_SIZE; i += 20) // 4 bytes is 2 pixels, so 20 bytes is every 10 pixels
	{
		if (i % WEBCAM_ROW_SIZE == 0)
		{
			if (i != 0)
			{
				i += (WEBCAM_ROW_SIZE * 9); // Increment by 9 so when adding we get to every 10th row
			}
			
			if (i == WEBCAM_TOTAL_SIZE)
			{
				break;
			}
		}
		
		GetSwitchRgbBytesFromYuyvAndSend(inputImage, i);
	}		
		
	DisableSwitch1ChipSelect();
}

// Gets bytes from a 640x480 image and converts to switch size
void LoadFramelessImage(char * inputImage)
{
	unsigned int horizontalCounter;
	unsigned int verticalCounter;
	unsigned int index;	
	
	DisableSwitchCommandSending(); // When commands are disabled data is able to be sent    
	EnableSwitch1ChipSelect();			
	
	index = 0;
	for (verticalCounter = 1; verticalCounter <= 64; verticalCounter++)
	{
		// To uniformly put 480 pixels into 64, every 2 lines we increase the index by 1 line
		if (verticalCounter != 0)
		{
			if (verticalCounter % 2 == 0)
			{
				index += WEBCAM_ROW_SIZE;
			}
			
			index += (WEBCAM_ROW_SIZE * 6); // 480/64 = every 6 rows
		}		
		
		for (horizontalCounter = 1; horizontalCounter <= 96; horizontalCounter++)
		{
			GetSwitchRgbBytesFromYuyvAndSend(inputImage, index);
			
			// To uniformly put 640 pixels into 96, every 3 pixels we increase the index by 2 pixels
			if (horizontalCounter % 3 == 0)
			{
				if (horizontalCounter != 0)
				{
					index += 4;
				}			
			}	
			
			index += 12; // 640 / 96 = 6 2/3 pixels, so increment by 12 bytes (2 bytes/ pixel)
		}
	}
		
	DisableSwitch1ChipSelect();
}

// Sends all setup parameters for this switch type
void ConfigureFramelessOledSwitch(void)
{
    unsigned char parameters[5] = {0};
    
    parameters[0] = FRAMELESS_CONTRAST_A_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_A, parameters, 1);
            
    parameters[0] = FRAMELESS_CONTRAST_B_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_B, parameters, 1);   
            
    parameters[0] = FRAMELESS_CONTRAST_C_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_C, parameters, 1);  
            
    parameters[0] = FRAMELESS_MASTER_CURRENT_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MASTER_CURRENT_CONTROL, parameters, 1);  
    
    parameters[0] = FRAMELESS_COLOR_A_PRECHARGE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_COLOR_A_PRECHARGE_SPEED, parameters, 1);  
    
    parameters[0] = FRAMELESS_COLOR_B_PRECHARGE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_COLOR_B_PRECHARGE_SPEED, parameters, 1);  
   
    parameters[0] = FRAMELESS_COLOR_C_PRECHARGE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_COLOR_C_PRECHARGE_SPEED, parameters, 1);  
            
    parameters[0] = FRAMELESS_COLOR_DEPTH_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_COLOR_DEPTH, parameters, 1);
    
    parameters[0] = FRAMELESS_DISPLAY_START_LINE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_START_LINE, parameters, 1);
    
    parameters[0] = FRAMELESS_DISPLAY_OFFSET_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_OFFSET, parameters, 1);
    
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_NORMAL, parameters, 0);
    
    parameters[0] = FRAMELESS_MULTIPLEX_RATIO_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MULTIPLEX_RATIO, parameters, 1);
    
    parameters[0] = FRAMELESS_DIM_MODE_DEFAULT_1;
    parameters[1] = FRAMELESS_DIM_MODE_DEFAULT_2;
    parameters[2] = FRAMELESS_DIM_MODE_DEFAULT_3;
    parameters[3] = FRAMELESS_DIM_MODE_DEFAULT_4;
    parameters[4] = FRAMELESS_DIM_MODE_DEFAULT_5;
    SendSwitchConfigurationCommand(SWITCH_COLOR_DIM_MODE, parameters, 5);
   
    parameters[0] = FRAMELESS_MASTER_CONFIGURATION_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MASTER_CONFIGURATION, parameters, 1);
    
    parameters[0] = FRAMELESS_POWER_SAVE_MODE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_POWER_SAVE_MODE, parameters, 1);
    
    parameters[0] = FRAMELESS_PHASE_PERIOD_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_PHASE_1_2_PERIOD_ADJUST, parameters, 1);
    
    parameters[0] = FRAMELESS_DISPLAY_CLOCK_DIVIDER_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_CLOCK_DIVIDER, parameters, 1);
    
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_ENABLE_LINEAR_GRAYSCALE, parameters, 0);
    
    parameters[0] = FRAMELESS_PRECHARGE_LEVEL_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_PRECHARGE_LEVEL, parameters, 1);
    
    parameters[0] = FRAMELESS_VCOMH_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_SET_VCOMH, parameters, 1);
  
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_ON, parameters, 0);
   
    parameters[0] = FRAMELESS_WINDOW_COLUMN_DEFAULT_1;
    parameters[1] = FRAMELESS_WINDOW_COLUMN_DEFAULT_2;
    SendSwitchConfigurationCommand(SWITCH_WINDOW_COLUMN, parameters, 2);
   
    parameters[0] = FRAMELESS_WINDOW_ROW_DEFAULT_1;
    parameters[1] = FRAMELESS_WINDOW_ROW_DEFAULT_2;
    SendSwitchConfigurationCommand(SWITCH_WINDOW_ROW, parameters, 2);    
}

// Sends all setup parameters for this switch type
void ConfigureStandardOledSwitch(void)
{
    unsigned char parameters[5] = {0};
    
    parameters[0] = STANDARD_CONTRAST_A_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_A, parameters, 1);
    
    parameters[0] = STANDARD_CONTRAST_B_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_B, parameters, 1);   
    
    parameters[0] = STANDARD_CONTRAST_C_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_CONTRAST_COLOR_C, parameters, 1); 
    
    parameters[0] = STANDARD_MASTER_CURRENT_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MASTER_CURRENT_CONTROL, parameters, 1);  
    
    parameters[0] = STANDARD_COLOR_DEPTH_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_COLOR_DEPTH, parameters, 1);
    
    parameters[0] = STANDARD_DISPLAY_START_LINE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_START_LINE, parameters, 1);
    
    parameters[0] = STANDARD_DISPLAY_OFFSET_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_OFFSET, parameters, 1);
    
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_NORMAL, parameters, 0);
    
    parameters[0] = STANDARD_MULTIPLEX_RATIO_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MULTIPLEX_RATIO, parameters, 1);
    
    parameters[0] = STANDARD_DIM_MODE_DEFAULT_1;
    parameters[1] = STANDARD_DIM_MODE_DEFAULT_2;
    parameters[2] = STANDARD_DIM_MODE_DEFAULT_3;
    parameters[3] = STANDARD_DIM_MODE_DEFAULT_4;
    parameters[4] = STANDARD_DIM_MODE_DEFAULT_5;
    SendSwitchConfigurationCommand(SWITCH_COLOR_DIM_MODE, parameters, 5);
    
    parameters[0] = STANDARD_MASTER_CONFIGURATION_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_MASTER_CONFIGURATION, parameters, 1);
    
    parameters[0] = STANDARD_POWER_SAVE_MODE_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_POWER_SAVE_MODE, parameters, 1);
    
    parameters[0] = STANDARD_PHASE_PERIOD_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_PHASE_1_2_PERIOD_ADJUST, parameters, 1);
    
    parameters[0] = STANDARD_DISPLAY_CLOCK_DIVIDER_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_CLOCK_DIVIDER, parameters, 1);
    
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_ENABLE_LINEAR_GRAYSCALE, parameters, 0);
  
    parameters[0] = STANDARD_PRECHARGE_LEVEL_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_PRECHARGE_LEVEL, parameters, 1);
    
    parameters[0] = STANDARD_VCOMH_DEFAULT;
    SendSwitchConfigurationCommand(SWITCH_SET_VCOMH, parameters, 1);
    
    
    parameters[0] = 0x00;
    SendSwitchConfigurationCommand(SWITCH_DISPLAY_ON, parameters, 0);
    
    parameters[0] = STANDARD_WINDOW_COLUMN_DEFAULT_1;
    parameters[1] = STANDARD_WINDOW_COLUMN_DEFAULT_2;
    SendSwitchConfigurationCommand(SWITCH_WINDOW_COLUMN, parameters, 2);
   
    parameters[0] = STANDARD_WINDOW_ROW_DEFAULT_1;
    parameters[1] = STANDARD_WINDOW_ROW_DEFAULT_2;
    SendSwitchConfigurationCommand(SWITCH_WINDOW_ROW, parameters, 2);           
}