/* 
 v1.0.41
 Reads frames from a webcam, converts the image to RGB and sends it to an NKK
 Standard or Frameless OLED switch. This uses the bcm2835 library, which required
 some operating system setup (such as turning on SPI). For getting the shutdown
 to work properly, a script has to be added to the /etc/rc.local file to run
 the program, with an & at the end so the program doesn't hang the operating system
 (IE /home/pi/NkkWebcam/RunNkkWebcamProgram.sh &).
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include "main.h"
#include "v4l2.h"
#include "SwitchFunctions.h"
#include "IoFunctions.h"

#define SWITCH_DEBOUNCE_TIME 250000
#define SHUTDOWN_COUNTER_MAX 20

char *webcamImage;

volatile bool _webcamImageRunning = true;
volatile bool _debounceTimerRunning = false;
struct itimerval _debouceTime;
typedef void (*sighandler_t)(int);

volatile uint _shutdownCounter = 0;
volatile bool _startShutdown = false;

void SetupDebounceTimer(void);
void StartTimer(void);
void ShutdownComputer(void);


int main(int argc, char** argv)
{    
	SetupRaspberryPiIo();

	SetupSwitch(); // Sends configuration options and sets blank image
	
	StartWebcamCapture();
	
	SetupDebounceTimer();
	
	while(1)
	{
		if (_startShutdown == true)
		{
			ShutdownComputer(); // Shuts down computer in main thread
		}
		
		if (_webcamImageRunning == true) // If they press the button we want the image frozen
		{
			webcamImage = GetWebcamFrame();
		}		
		
		LoadWebcamImageOnSwitch(webcamImage);
		
		bool pressed = ReadSwitchPress();
		if (pressed == true)
		{
			if (_debounceTimerRunning == false)
			{
				_webcamImageRunning = !_webcamImageRunning; // Freeze the image on the switch if the button is pressed	
				StartTimer();
			}	
		}
		else
		{
			_shutdownCounter = 0;
		}
	}			
    
    return 0;
}

// Callback from the system that the button debounce timer has expired
void TimerExpired(int nothing)
{
	_debounceTimerRunning = false;
	_shutdownCounter++;
	if (_shutdownCounter >= SHUTDOWN_COUNTER_MAX) // If there are SHUTDOWN_COUNTER_MAX consecutive button presses we shut down the system
	{
		_startShutdown = true;
	}
}

// Provides all setup for the debounce timer
void SetupDebounceTimer(void)
{
	_debouceTime.it_interval.tv_sec = 0;
	_debouceTime.it_interval.tv_usec = 0;
	_debouceTime.it_value.tv_sec = 0;
	_debouceTime.it_value.tv_usec = 500000;		
	
	struct sigaction debounceSetup;
	debounceSetup.sa_handler = &TimerExpired;	
	sigaction(SIGALRM, &debounceSetup,NULL); // Tells the system the timer length and the callback function for the timer expired
}

// Starts the debounce timer
void StartTimer(void)
{
	_debounceTimerRunning = true;		
	setitimer(ITIMER_REAL, &_debouceTime, 0); // Starts the timer
}

// Shuts down the computer safely
void ShutdownComputer(void)
{
	FinishWebcamCapture();
    SafelyCloseIoFunctions();
	
	system("shutdown -P now");
	exit(EXIT_SUCCESS);
}

