/*
 *  This code is based on https://gist.github.com/maxlapshin/1253534 . However,
 *  that code is extremely poorly commented and very confusing to read. It
 *	has been re-written for this purpose, hopefully easier to follow and understand. The
 *	API descriptions for all the commands and enums is located at: 
 *	https://linuxtv.org/downloads/v4l-dvb-apis-new/uapi/v4l/v4l2.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <getopt.h>             /* getopt_long() */

#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <linux/videodev2.h>
#include "v4l2.h"

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#ifndef V4L2_PIX_FMT_H264
#define V4L2_PIX_FMT_H264     v4l2_fourcc('H', '2', '6', '4') /* H264 with start codes */
#endif

struct bufferStruct 
{
	void   *start; // The starting address of the data for the buffer
	size_t  length; // Length of the data in the buffer
};

static char            *linuxDeviceName;
static int              fileDescriptor = -1;
struct bufferStruct		*deviceOutputBuffers;
static unsigned int     numberOfBuffers;

char * finalImage; // Image to return after getting from webcam and converting to RGB

static void PrintErrorAndExit(const char *errorString)
{
        fprintf(stderr, "%s error %d, %s\n", errorString, errno, strerror(errno));
        exit(EXIT_FAILURE);
}

// Used to query or command V4L2 devices
static int QueryCommandDevice(int fileDescriptor, int request, void *arguments)
{
        int r;

        do 
		{
			r = ioctl(fileDescriptor, request, arguments);
        } 
		while (-1 == r && EINTR == errno);

        return r;
}

static void process_image(const void *p, int size)
{
	finalImage = p; // We assign our "global" variable to the read memory address and return it for processing
}

static int read_frame(void)
{
        struct v4l2_buffer buf;

		CLEAR(buf);

		buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		buf.memory = V4L2_MEMORY_MMAP;

		// VIDIOC_DQBUF reads from the device memory.
		if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_DQBUF, &buf)) 
		{
			switch (errno) 
			{
			case EAGAIN:
				return 0;

			case EIO:
				/* Could ignore EIO, see spec. */

				/* fall through */

			default:
				PrintErrorAndExit("VIDIOC_DQBUF");
			}
		}

		assert(buf.index < numberOfBuffers);

		process_image(deviceOutputBuffers[buf.index].start, buf.bytesused);

		// VIDIOC_QBUF tells the device to fill the next buffer with the next frame.
		if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_QBUF, &buf))
				PrintErrorAndExit("VIDIOC_QBUF");

        return 1;
}

static void SetupAndGetFrame(void)
{
        fd_set fileDescriptorSet; // A set(array?) of file descriptors
//		struct timeval tv; // Some webcams take a while so we disable timeout
		int r;

		FD_ZERO(&fileDescriptorSet); // Clears the set
		FD_SET(fileDescriptor, &fileDescriptorSet); // Adds our file descriptor to the empty set

		/* Timeout. */
//		tv.tv_sec = 2;
//		tv.tv_usec = 0;

		/*	This function blocks until the device has filled the buffer with data.
			Pay attention to the device limitiations as that will affect the amount
			of time spent in this function (IE some webcams are hard limited
			to 30 frames per second) 
		 */
		/*	Per linux manual: int select(int nfds, fd_set *readfds, fd_set *writefds,
           fd_set *exceptfds, struct timeval *timeout);
			Allows a program to monitor multiple file descriptors, 
			waiting until one or more of the file descriptors become "ready" for some class of I/O operation.
			*nfds is the highest-numbered file descriptor in any of the three sets, plus 1.
		 
		 */
		r = select(fileDescriptor + 1, &fileDescriptorSet, NULL, NULL, NULL);

		if (-1 == r) 
		{
			if (EINTR != errno)					
				PrintErrorAndExit("select");
		}

		if (0 == r) 
		{
			fprintf(stderr, "select timeout\n");
			exit(EXIT_FAILURE);
		}

		read_frame();
}

static void stop_capturing(void)
{
        enum v4l2_buf_type type;

		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_STREAMOFF, &type))
			PrintErrorAndExit("VIDIOC_STREAMOFF");
}

static void start_capturing(void)
{
        unsigned int i;
        enum v4l2_buf_type type;

         for (i = 0; i < numberOfBuffers; ++i) 
		 {
			struct v4l2_buffer buf;

			CLEAR(buf);
			buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory = V4L2_MEMORY_MMAP;
			buf.index = i;

			if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_QBUF, &buf))
					PrintErrorAndExit("VIDIOC_QBUF");
		 }
		
		type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_STREAMON, &type))
				PrintErrorAndExit("VIDIOC_STREAMON");
}

static void uninit_device(void)
{
        unsigned int i;

		for (i = 0; i < numberOfBuffers; ++i)
		{
			if (-1 == munmap(deviceOutputBuffers[i].start, deviceOutputBuffers[i].length))
				PrintErrorAndExit("munmap");
		}		

        free(deviceOutputBuffers);
}

static void init_mmap(void)
{
        struct v4l2_requestbuffers req;

        CLEAR(req);

        req.count = 4;
        req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        req.memory = V4L2_MEMORY_MMAP;

        if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_REQBUFS, &req)) 
		{
			if (EINVAL == errno) 
			{
				fprintf(stderr, "%s does not support "
						 "memory mapping\n", linuxDeviceName);
				exit(EXIT_FAILURE);
			} 
			else 
			{
				PrintErrorAndExit("VIDIOC_REQBUFS");
			}
        }

        if (req.count < 2) 
		{
			fprintf(stderr, "Insufficient buffer memory on %s\n",
					 linuxDeviceName);
			exit(EXIT_FAILURE);
        }

        deviceOutputBuffers = calloc(req.count, sizeof(*deviceOutputBuffers));

        if (!deviceOutputBuffers) 
		{
			fprintf(stderr, "Out of memory\n");
			exit(EXIT_FAILURE);
        }

        for (numberOfBuffers = 0; numberOfBuffers < req.count; ++numberOfBuffers) 
		{
			struct v4l2_buffer buf;

			CLEAR(buf);

			buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			buf.memory      = V4L2_MEMORY_MMAP;
			buf.index       = numberOfBuffers;

			if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_QUERYBUF, &buf))
					PrintErrorAndExit("VIDIOC_QUERYBUF");

			// We map the memory returned by the command to our local struct
			deviceOutputBuffers[numberOfBuffers].length = buf.length;
			deviceOutputBuffers[numberOfBuffers].start =
					mmap(NULL /* start anywhere */,
						  buf.length,
						  PROT_READ | PROT_WRITE /* required */,
						  MAP_SHARED /* recommended */,
						  fileDescriptor, buf.m.offset);

			if (MAP_FAILED == deviceOutputBuffers[numberOfBuffers].start)
					PrintErrorAndExit("mmap");
        }
}

static void init_device(void)
{
        struct v4l2_capability deviceCapabilities;
        struct v4l2_cropcap deviceCroppingCapabilities;
        struct v4l2_crop currentCropSettings;
        struct v4l2_format devicePixelFormat;
        unsigned int minimumImageBytes;

		// Gets the capabilities of the device
        if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_QUERYCAP, &deviceCapabilities)) 
		{
			if (EINVAL == errno) 
			{
					fprintf(stderr, "%s is no V4L2 device\n",
                                 linuxDeviceName);
					exit(EXIT_FAILURE);
			} 
			else 
			{
					PrintErrorAndExit("VIDIOC_QUERYCAP");
			}
        }

		// Error catching if there is nothing to get
        if (!(deviceCapabilities.capabilities & V4L2_CAP_VIDEO_CAPTURE)) 
		{
			fprintf(stderr, "%s is no video capture device\n",
					 linuxDeviceName);
			exit(EXIT_FAILURE);
        }

		// Checks that the device can actually get video in the way requested
		if (!(deviceCapabilities.capabilities & V4L2_CAP_STREAMING)) 
		{
			fprintf(stderr, "%s does not support streaming i/o\n",
					 linuxDeviceName);
			exit(EXIT_FAILURE);
		}


        /* Select video input, video standard and tune here. */

        CLEAR(deviceCroppingCapabilities);

        deviceCroppingCapabilities.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

		// Queries the cropping limits and pixel aspect to calculate scaling factors
        if (0 == QueryCommandDevice(fileDescriptor, VIDIOC_CROPCAP, &deviceCroppingCapabilities)) 
		{
			currentCropSettings.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
			currentCropSettings.c = deviceCroppingCapabilities.defrect; /* reset to default */

			if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_S_CROP, &currentCropSettings)) 
			{
					switch (errno) 
					{
					case EINVAL:
							/* Cropping not supported. */
							break;
					default:
							/* Errors ignored. */
							break;
					}
			}
        } 
		else 
		{
                /* Errors ignored. */
        }


		/*	Gets the device pixel format
		 *	This device doesn't seem to support setting specific outputs,
		 *	it will only output in YUYV */
		
        CLEAR(devicePixelFormat);

		devicePixelFormat.type = V4L2_BUF_TYPE_VIDEO_CAPTURE; // We set the type we want to query for
		QueryCommandDevice(fileDescriptor, VIDIOC_G_FMT, &devicePixelFormat); // Queries the device and returns the pixel format
        
		/*	Initially attempted to set the webcam settings but this webcam does not let us, always returns 640x480 YUYV 
		 *	Keeping the code here in case webcam changes
		 */
		
//		devicePixelFormat.fmt.pix.width       = 640;
//		devicePixelFormat.fmt.pix.height      = 480; 
		//fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_H264; //replace
//		devicePixelFormat.fmt.pix.pixelformat = V4L2_PIX_FMT_RGB24; // TODO:check if this is correct
//		devicePixelFormat.fmt.pix.field       = V4L2_FIELD_NONE;
//		fmt.fmt.pix.bytesperline = 1920;
//		fmt.fmt.pix.sizeimage = 921600;
//		fmt.fmt.pix.colorspace = V4L2_COLORSPACE_SRGB;
//		fmt.fmt.pix.priv = 0;

		if (-1 == QueryCommandDevice(fileDescriptor, VIDIOC_S_FMT, &devicePixelFormat))
			PrintErrorAndExit("VIDIOC_S_FMT");

        /* Buggy driver paranoia. */
        minimumImageBytes = devicePixelFormat.fmt.pix.width * 2;
        if (devicePixelFormat.fmt.pix.bytesperline < minimumImageBytes)
                devicePixelFormat.fmt.pix.bytesperline = minimumImageBytes;
        minimumImageBytes = devicePixelFormat.fmt.pix.bytesperline * devicePixelFormat.fmt.pix.height;
        if (devicePixelFormat.fmt.pix.sizeimage < minimumImageBytes)
                devicePixelFormat.fmt.pix.sizeimage = minimumImageBytes;

        init_mmap();
}

static void close_device(void)
{
        if (-1 == close(fileDescriptor))
                PrintErrorAndExit("close");

        fileDescriptor = -1;
}

// Gets the information for the linux device specified and puts it in a struct
static void open_device(void)
{
        struct stat deviceInformation; // Struct containing device information

		// Gets the device info and puts it in the deviceInformation struct
        if (-1 == stat(linuxDeviceName, &deviceInformation)) 
		{
			fprintf(stderr, "Cannot identify '%s': %d, %s\n",
					 linuxDeviceName, errno, strerror(errno));
			exit(EXIT_FAILURE);
        }

		// Checks if a device is present
        if (!S_ISCHR(deviceInformation.st_mode)) 
		{
			fprintf(stderr, "%s is no device\n", linuxDeviceName);
			exit(EXIT_FAILURE);
        }

		
        fileDescriptor = open(linuxDeviceName, O_RDWR /* required */ | O_NONBLOCK, 0);

        if (-1 == fileDescriptor) // Error catching
		{
			fprintf(stderr, "Cannot open '%s': %d, %s\n",
					 linuxDeviceName, errno, strerror(errno));
			exit(EXIT_FAILURE);
        }
}

// Does all setup and tells the webcam to start acquiring images
void StartWebcamCapture(void)
{
	linuxDeviceName = "/dev/video0"; // Set the default device name
	open_device();
	init_device();
	start_capturing();
}

// Tells the webcam to stop acquiring and cleans up setup
void FinishWebcamCapture(void)
{
	stop_capturing();
	uninit_device();
	close_device();
	fprintf(stderr, "\n");
}

char * GetWebcamFrame(void)
{
	SetupAndGetFrame();
	
	return finalImage;
}